"""SDV-Installer command line interface."""

import argparse
import sys

from sdv_installer.installation import (
    download_packages,
    get_auth_file,
    install_packages,
    install_packages_from_folder,
    list_packages,
    uninstall_packages,
)
from sdv_installer.utils.system_requirements import print_check_results

# Adding a new version here to avoid circular imports
SDV_INSTALLER_VERSION = '0.4.0'


def install_action(args):
    """Install the `SDV-Enterprise` packages."""
    if args.folder:
        install_packages_from_folder(
            folder=args.folder,
            package=args.package,
            options=args.options,
            debug=args.debug,
        )

    else:
        install_packages(
            username=args.username,
            license_key=args.license_key,
            credentials=args.credentials,
            package=args.package,
            options=args.options,
            debug=args.debug,
            version=args.version,
            upgrade=args.upgrade,
        )


def uninstall_action(args):
    """Uninstall the `SDV-Enterprise` packages."""
    uninstall_packages(debug=args.debug)


def download_action(args):
    """Download SDV-Enterprise packages for offline install."""
    download_packages(
        username=args.username,
        license_key=args.license_key,
        credentials=args.credentials,
        package=args.package,
        python_version=args.python_version,
        platform=args.platform,
        folder=args.folder,
        version=args.version,
        options=args.options,
        debug=args.debug,
    )


def list_action(args, parser=None):
    """List all packages the user has access to."""
    list_packages(
        username=args.username, license_key=args.license_key, credentials=args.credentials
    )


def authenticate_action(args):
    """Download auth file."""
    get_auth_file(
        username=args.username,
        license_key=args.license_key,
        folder=args.folder,
        credentials=args.credentials,
    )


def perform_check_action(args):
    """Perform a check on the user's system to verify compatibility with SDV-Enterprise."""
    print_check_results()


def get_parser():
    """Return the argparser for the SDV-Installer application."""
    parser = argparse.ArgumentParser(description='SDV Installer.')
    parser.set_defaults(action=None)

    # Top-level version flag
    parser.add_argument('--version', action='version', version=SDV_INSTALLER_VERSION)
    action = parser.add_subparsers(title='action')

    # Install
    install_parser = action.add_parser('install', help='Install your SDV Enterprise Packages.')
    install_parser.set_defaults(action=install_action)
    install_parser.add_argument(
        '--package',
        required=False,
        help='The package that you want to install.',
    )
    install_parser.add_argument(
        '--options', help='Optional dependencies to install.', required=False, nargs='+'
    )
    install_parser.add_argument('--folder', type=str, help='Install from a local folder.')
    install_parser.add_argument('--debug', action='store_true', help='Print pip install commands.')
    install_parser.add_argument(
        '-v', '--version', required=False, help='SDV-Enterprise version (e.g., 0.23.0).'
    )
    install_parser.add_argument(
        '-u',
        '--upgrade',
        required=False,
        help='Upgrade the current installed packages.',
        action='store_true',
    )
    install_parser.add_argument('--username', required=False, help='SDV-Enterprise username.')
    install_parser.add_argument('--license-key', required=False, help='SDV-Enterprise license key.')
    install_parser.add_argument(
        '--credentials', required=False, help='Path to a credentials json file.'
    )

    # Uninstall
    uninstall_parser = action.add_parser(
        'uninstall', help='Uninstall your SDV Enterprise Packages.'
    )
    uninstall_parser.set_defaults(action=uninstall_action)
    uninstall_parser.add_argument(
        '--debug', action='store_true', help='Print pip install commands.'
    )

    # Download
    download_parser = action.add_parser(
        'download',
        help='Download your SDV Enterprise Packages for offline installation.',
    )
    download_parser.set_defaults(action=download_action)
    download_parser.add_argument(
        '--package',
        required=False,
        help=(
            'The package that you want to download. If not provided, all '
            'packages will be downloaded'
        ),
    )
    download_parser.add_argument(
        '--python-version', required=False, help='Python version (e.g., 39).'
    )
    download_parser.add_argument(
        '--platform',
        required=False,
        help='Platform tag(s) (e.g., macosx_14_0_arm64). You can specify this multiple times.',
        action='append',
    )
    download_parser.add_argument(
        '--folder', required=True, help='Destination folder for downloaded packages.'
    )
    download_parser.add_argument(
        '--debug', action='store_true', help='Print pip download commands.'
    )
    download_parser.add_argument(
        '--version', required=False, help='SDV-Enterprise version (e.g., 0.23).'
    )
    download_parser.add_argument(
        '--options', help='Optional dependencies to install.', required=False, nargs='+'
    )
    download_parser.add_argument('--username', required=False, help='SDV-Enterprise username.')
    download_parser.add_argument(
        '--license-key', required=False, help='SDV-Enterprise license key.'
    )
    download_parser.add_argument(
        '--credentials', required=False, help='Path to a credentials json file.'
    )

    # List
    list_parser = action.add_parser('list-packages', help='List all packages you have access to.')
    list_parser.set_defaults(action=list_action)
    list_parser.add_argument('--username', required=False, help='SDV-Enterprise username.')
    list_parser.add_argument('--license-key', required=False, help='SDV-Enterprise license key.')
    list_parser.add_argument(
        '--credentials', required=False, help='Path to a credentials json file.'
    )

    # Authenticate
    auth_parser = action.add_parser(
        'authenticate', help='Download auth file to required destination or specified folder.'
    )
    auth_parser.set_defaults(action=authenticate_action)
    auth_parser.add_argument('--username', required=False, help='SDV-Enterprise username.')
    auth_parser.add_argument('--license-key', required=False, help='SDV-Enterprise license key.')
    auth_parser.add_argument('--folder', required=False, help='Destination folder for auth file.')
    auth_parser.add_argument(
        '--credentials', required=False, help='Path to a credentials json file.'
    )

    # Check Requirements
    check_parser = action.add_parser(
        'check-requirements', help='Check if your system meets SDV Enterprise requirements.'
    )
    check_parser.set_defaults(action=perform_check_action)

    return parser


def run_cli():
    """Run command line interface."""
    parser = get_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()
    args.action(args)


if __name__ == '__main__':
    run_cli()



"""Licensing authentication module."""

import json
from functools import wraps

import requests
from requests.exceptions import RequestException

from sdv_installer import config
from sdv_installer.utils import (
    get_password_input,
    print_empty_username_or_license_key,
    print_failed_to_connect,
    print_invalid_credentials,
)


def authenticate_user_and_license_key(username, license_key):
    """Authenticate the user and license key by sending a request to the authentication API.

    Constructs an authentication request using the provided username and license key,
    sends it to the appropriate API endpoint, and returns a boolean indicating whether
    the authentication was successful.

    Args:
        username (str):
            The username to authenticate.
        license_key (str):
            The license key to authenticate.

    Returns:
        bool:
            True if authentication is successful, False otherwise.

    Raises:
        InvalidLoginCredentials:
            If the user has provided invalid login credentials.
    """
    post_data = {'username': username, 'license_key': license_key}
    try:
        response = requests.post(
            config.API_VALIDATE, json=post_data, headers=config.HEADERS, timeout=config.TIMEOUT
        )
    except RequestException:
        print_failed_to_connect()
        return False

    response_dict = response.json()
    valid = response_dict.get('valid', False)
    authenticated = bool(response.status_code == requests.codes.ok and valid)
    if not authenticated:
        print_invalid_credentials()

    return authenticated


def _load_credentials_from_file(filepath):
    try:
        with open(filepath, 'r') as file:
            credentials = json.load(file)
            return credentials

    except FileNotFoundError as e:
        raise ValueError(f'The credentials file {filepath} was not found.') from e
    except json.JSONDecodeError as e:
        raise RuntimeError(f'Failed to decode the json file {filepath}') from e


def authenticate(function):
    """Authenticate wrapper to be used by other functions."""

    @wraps(function)
    def wrapper(*args, **kwargs):
        username = kwargs.get('username')
        license_key = kwargs.get('license_key')
        credentials = kwargs.get('credentials')
        if (username or license_key) and credentials:
            raise ValueError(
                'You have provided both --credentials and --username/--license-key flags. '
                'Please choose only one of these methods for authentication.'
            )

        elif username or license_key:
            username_no_lk = username is not None and license_key is None
            lk_no_username = license_key is not None and username is None
            if username_no_lk or lk_no_username:
                raise ValueError(
                    'You must provide both the --username and --license-key flags for '
                    'authentication.'
                )

        elif credentials:
            credentials_json = _load_credentials_from_file(credentials)
            username = credentials_json['username']
            license_key = credentials_json['license_key']
            kwargs['username'] = username
            kwargs['license_key'] = license_key

        else:
            username = input('Username: ') or ''
            username = username.strip()

            license_key = get_password_input('License Key: ') or ''
            license_key = license_key.strip()

            if not username or not license_key:
                print_empty_username_or_license_key()
                return

            kwargs['username'] = username
            kwargs['license_key'] = license_key

        authenticated = authenticate_user_and_license_key(username, license_key)
        if authenticated:
            return function(*args, **kwargs)

    return wrapper



"""Constants used across the SDV-Installer."""

from enum import IntEnum

try:
    from enum import StrEnum as StringEnum

except ImportError:
    from enum import Enum as StringEnum

from packaging.version import Version

from sdv_installer.config import PYPI_URL

# System Requirement Constants

MIN_PYTHON_VERSION = Version('3.9')
MAX_PYTHON_VERSION = Version('3.14')
REQUIRED_BITNESS = '64-bit'
AUTH_FILE_NAME = 'auth.json'

SUPPORTED_PLATFORM_TAGS = [
    'macosx_10_9_x86_64',
    'macosx_11_0_universal2',
    'macosx_11_0_x86_64',
    'macosx_12_0_arm64',
    'macosx_12_0_x86_64',
    'macosx_13_0_arm64',
    'macosx_13_0_x86_64',
    'macosx_14_0_arm64',
    'macosx_14_0_x86_64',
    'manylinux2010_x86_64',
    'manylinux2014_x86_64',
    'manylinux_2_17_x86_64',
    'manylinux_2_28_x86_64',
    'musllinux_1_1_x86_64',
    'win_amd64',
]
SUPPORTED_OSES = ['macOS', 'Linux', 'Windows']
SUPPORTED_ARCHITECTURES = ['x86_64', 'arm64', 'amd64', 'win-amd64']
MIN_PIP_VERSION = Version('22.3')
MIN_OS_VERSIONS = {
    'macOS': Version('11'),
    'windows': Version('10.0.10240'),
    'linux': Version('5.10'),
}
SUPPORTED_LINUX_DISTROS = {
    'debian': Version('11'),
    'ubuntu': Version('22.04'),
}

# Packaging
EXPECTED_PACKAGES = set([
    'sdv-enterprise',
    'bundle-cag',
    'bundle-ai-connectors',
    'bundle-xsynthesizers',
    'bundle-differential-privacy',
])

TRUSTED_HOSTS = [
    'pypi.org',
    'pypi.python.org',
    'files.pythonhosted.org',
    PYPI_URL,
]

CONNECTORS_PACKAGES = [
    'bundle-ai-connectors',
]

# Message Constants
ACTION_SUCCESS_MESSAGE = {
    'install': 'Installed!',
    'uninstall -y': 'Uninstalled!',
    'download': 'Downloaded!',
}

ACTION_SUCCESS_ALREADY_INSTALLED = {
    'install': 'N/A (already installed)',
}

ACTION_SUCCESS_ALREADY_UPDATED = {
    'install': 'N/A (already installed and up-to-date)',
}

ACTION_ERROR_MESSAGE = {
    'install': 'Installation failed. Use --debug for more details.',
    'uninstall -y': 'Uninstall failed. Use --debug for more details.',
    'download': 'Download failed. Use --debug for more details.',
}
ADDITIONAL_DEPS_MESSAGE_PREFIX = (
    'Note: This package has additional required dependencies that were also installed'
)

# Constants for keypresses
ENTER_KEY = '\r'
BACKSPACE_KEYS = ['\x08', '\x7f']
NEWLINE_KEY = '\n'

# Constants for Summary Messages
SUMMARY_SUCCESS_MESSAGE = {
    'install': ('Success! All packages have been installed. You are ready to use SDV Enterprise.'),
    'download': (
        'Success! All packages have been downloaded. '
        'Please install the packages to use SDV Enterprise.'
    ),
}

SUMMARY_PARTIAL_SUCCESS_MESSAGE = {
    'install': (
        'Notice! All packages were installed, but `sdv-enterprise` was not included.\n'
        'To use SDV Enterprise, please make sure it is installed.'
    ),
    'download': (
        'Notice! All packages were downloaded, but `sdv-enterprise` was not included.\n'
        'Please make sure that you have downloaded SDV-Enterprise in order to '
        'be able to install it later.'
    ),
}

SUMMARY_WARNING_MESSAGE = {
    'install': (
        'Warning! Some packages that you have access to could not be installed. '
        'Please check your setup and contact DataCebo to troubleshoot.'
    ),
    'download': (
        'Warning! Some packages that you have access to could not be downloaded. '
        'Please check your setup and contact DataCebo to troubleshoot.'
    ),
}


class ProductType(StringEnum):
    """Product type enum."""

    BASE = 'base'
    BUNDLE = 'bundle'


class ExitCode(IntEnum):
    """Exit Code Representation."""

    SUCCESS = 0
    ERROR = 1



"""Installation Module."""

from sdv_installer.installation.installer import (
    download_packages,
    get_auth_file,
    install_packages_from_folder,
    install_packages,
    list_packages,
    uninstall_packages,
)

__all__ = (
    'download_packages',
    'get_auth_file',
    'install_packages',
    'install_packages_from_folder',
    'list_packages',
    'uninstall_packages',
)



"""Configuration module for SDV-Installer."""

import os

# Default API URL that will be used
LKS_URL = os.getenv('LKS_URL', 'https://lks.datacebo.com')

# Endpoints
API_PACKAGE_PERMISSIONS = f'{LKS_URL}/api/v1/package_permissions/active'
API_VALIDATE = f'{LKS_URL}/api/v1/licenses/validate'
API_AUTH_FILE = f'{LKS_URL}/api/v1/licenses/auth_file'

# Pypi Server to be used
PYPI_URL = os.getenv('PYPI_URL', 'pypi.datacebo.com')

# Headers
HEADERS = {'accept': 'application/json', 'Content-Type': 'application/json'}

# Timeout Value for API Calls
TIMEOUT = os.getenv('SDV_API_TIMEOUT', 10)



"""Authentication module."""

from sdv_installer.authentication.authentication import (
    authenticate,
    authenticate_user_and_license_key,
)

__all__ = ('authenticate', 'authenticate_user_and_license_key')



"""Utility functions to handle requests HTTP errors."""


class NotFoundError(ValueError):
    """Error to be raised when an end point is not found (404)."""


def handle_http_error_response(response):
    """Handle HTTP error responses and raises appropriate exceptions with error messages.

    The function attempts to parse the error details from the response as JSON. If the content
    cannot be parsed as JSON, the raw response text will be used.

    Args:
        response (requests.Response):
            The HTTP response object that contains the status code and error details.

    Raises:
        ValueError:
            If the status code is 400 (Bad Request) or 422 (Validation Error).
        PermissionError:
            If the status code is 401 (Unauthorized) or 403 (Forbidden).
        NotFoundError:
            If the status code is 404 (Not Found).
        RuntimeError:
            If the status code is 405 (Method Not Allowed) or 5xx (Server Error).
        Exception:
            For any unexpected status code.
    """
    status = response.status_code
    try:
        error_detail = response.json()
    except ValueError:
        error_detail = response.text

    error_map = {
        400: (ValueError, f'Bad Request: {error_detail}'),
        401: (PermissionError, f'Invalid Credentials or Expired License Key: {error_detail}.'),
        403: (PermissionError, 'Forbidden: You don’t have permission to access this resource.'),
        404: (NotFoundError, 'Not Found: The requested endpoint does not exist.'),
        405: (RuntimeError, 'Method Not Allowed: Check the HTTP method.'),
        422: (ValueError, f'Validation Error: {error_detail}'),
    }

    if status >= 500:
        exception, message = RuntimeError, f'Server Error ({status}): {error_detail}'

    else:
        exception, message = error_map.get(
            status, (Exception, f'Unexpected Error ({status}): {error_detail}')
        )

    raise exception(message)



"""Utility functions for securely handling password input with character masking."""

import sys
import time
from urllib.parse import quote_plus, urlparse, urlunparse

from sdv_installer.config import API_VALIDATE, PYPI_URL
from sdv_installer.constants import (
    ACTION_ERROR_MESSAGE,
    ACTION_SUCCESS_ALREADY_INSTALLED,
    ACTION_SUCCESS_ALREADY_UPDATED,
    ACTION_SUCCESS_MESSAGE,
    ADDITIONAL_DEPS_MESSAGE_PREFIX,
    BACKSPACE_KEYS,
    ENTER_KEY,
    NEWLINE_KEY,
    SUMMARY_PARTIAL_SUCCESS_MESSAGE,
    SUMMARY_SUCCESS_MESSAGE,
    SUMMARY_WARNING_MESSAGE,
    ExitCode,
)
from sdv_installer.utils.package_utils import is_version_bigger, is_version_equal


def print_with_flush(text):
    """Helper function to print text and flush the output immediately."""
    sys.stdout.write(text)
    sys.stdout.flush()


def handle_keypress(character, password):
    """Handles user keypresses: backspace and normal characteracter input."""
    if character in BACKSPACE_KEYS:
        if password:
            password = password[:-1]
            print_with_flush('\b \b')

    else:
        password += character
        print_with_flush('*')

    return password


def get_password(prompt, get_char_func):
    """Generic password input handler."""
    print_with_flush(prompt)
    password = ''
    while True:
        char = get_char_func()
        char = char.decode('utf-8') if isinstance(char, bytes) else char
        if char in [ENTER_KEY, NEWLINE_KEY]:
            print_with_flush('\n')
            break
        password = handle_keypress(char, password)

    return password


def get_char_win():
    """Windows-specific character reading function."""
    import msvcrt

    return msvcrt.getch()


def get_char_unix():
    """Unix-specific character reading function."""
    import termios
    import tty

    fd = sys.stdin.fileno()
    old_settings = termios.tcgetattr(fd)
    try:
        tty.setraw(fd)
        return sys.stdin.read(1)
    finally:
        termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)


def get_password_input(prompt='License Key: '):
    """Password getter with stars."""
    if sys.platform == 'win32':
        return get_password(prompt, get_char_win)

    return get_password(prompt, get_char_unix)


def print_message(message):
    """Function to be used to print messages across the library."""
    print(message)  # noqa


def handle_process_exit_code(
    process,
    package,
    printable_package,
    version=None,
    action=None,
    installed_packages=None,
    upgrade=False,
):
    """Handle and display messages based on a subprocess exit code.

    Determines the appropriate success or error message to display
    based on the exit code of a subprocess and the package installation context.

    Args:
        process (subprocess.Process):
            A completed subprocess with a `returncode` attribute.
        package (str):
            The internal package name.
        printable_package (str):
            The user-friendly package name for display.
        version (str, optional):
            The target version of the package. Defaults to None.
        action (str, optional):
            The action performed (e.g., 'install', 'upgrade'). Defaults to None.
        installed_packages (dict, optional):
            A mapping of package names to installed versions. Defaults to None.
        upgrade (bool, optional):
            Indicates whether this is an upgrade operation. Defaults to False.
    """
    if process.returncode == ExitCode.SUCCESS:
        msg = ACTION_SUCCESS_MESSAGE.get(action)
        if action == 'install':
            installed_version = installed_packages.get(package) if installed_packages else None
            if installed_version:
                is_installed_bigger = is_version_bigger(installed_version, version)
                is_installed_equal = is_version_equal(installed_version, version)
                if upgrade and is_installed_bigger or is_installed_equal:
                    msg = ACTION_SUCCESS_ALREADY_UPDATED.get(action)

                elif installed_version and is_installed_equal:
                    msg = ACTION_SUCCESS_ALREADY_INSTALLED.get(action)

    elif process.returncode >= ExitCode.ERROR:
        msg = ACTION_ERROR_MESSAGE.get(action)

    sys.stdout.write(f'\r{printable_package} - {msg}\n')
    sys.stdout.flush()


def display_progress_animation(
    process,
    package,
    version=None,
    action=None,
    installed_packages=None,
    upgrade=False,
    show_version=False,
):
    """Display an animated progress indicator while a subprocess is running.

    This function continuously displays an animated progress message
    (e.g., "package .", "package ..", "package ...") until the given
    subprocess finishes. Once completed, it prints a final message
    based on the process's exit code.

    Args:
        process (subprocess.Popen):
            The subprocess to monitor.
        package (str):
            The name of the package being processed.
        version (str, optional):
            An optional version string to display with the package name.
        action (str, optional):
            The action being performed (e.g., "install", "download"). Used to determine
            the appropriate success or error message.
    """
    printable_package = f'{package} (version {version})' if show_version and version else package

    frames = [f'{printable_package} .  ', f'{printable_package} .. ', f'{printable_package} ...']
    while True:
        for frame in frames:
            sys.stdout.write(f'\r{frame}')
            sys.stdout.flush()
            time.sleep(0.3)
            if process.poll() is not None:
                handle_process_exit_code(
                    process,
                    package,
                    printable_package,
                    version,
                    action,
                    installed_packages,
                    upgrade,
                )
                return


def print_invalid_credentials():
    """Print an error message for invalid or expired SDV Enterprise credentials."""
    print_message(
        'Error installing SDV Enterprise. This may be due to an invalid '
        'username/license key combo, or because your license key has expired. '
        "Please double-check that you're providing the right credentials. If "
        "you're continuing to experience issues, please reach out to the "
        'DataCebo team. '
    )


def print_failed_to_connect():
    """Print an error message indicating failure to connect to the authentication server."""
    print_message(
        'Failed to connect to the authentication server.\n'
        'Please check your internet connection or firewall settings and that the API server '
        f'{API_VALIDATE} is reachable.'
    )


def print_warning_base_connector_package_installed():
    """Print a 'warning' message indicating that the base connector package will be installed."""
    print_message(
        "\nWarning: In order to use AI Connectors, you'll need to install "
        'database-specific packages. Please use the `--options` '
        'flag to specify your database variants.'
    )


def print_additional_dependencies_installed(additional_dependencies):
    """Print a message indicating that additional enterprise-related packages were installed.

    Args:
        action (list[str]):
            The list of additional packages that were installed.
    """
    if len(additional_dependencies) == 0:
        return

    message = f'{ADDITIONAL_DEPS_MESSAGE_PREFIX} ('
    message += ', '.join(additional_dependencies) + ').'
    print_message('\n' + message)


def print_package_summary(action, package_status, sdv_included):
    """Print a summary message indicating that whether or not the installation was successful.

    Args:
        action (str):
            The action that was performed (download or install).
        package_status (dict):
            A dictionary with the packages and whether or not the action was successfully performed.
        sdv_included (bool):
            Whether sdv-enterprise was included in the successful result.
    """
    all_success = all(package_status.values())

    if all_success and sdv_included:
        print_message('\n' + SUMMARY_SUCCESS_MESSAGE[action])

    elif all_success and not sdv_included:
        print_message('\n' + SUMMARY_PARTIAL_SUCCESS_MESSAGE[action])
    else:
        print_message('\n' + SUMMARY_WARNING_MESSAGE[action])


def mask_license_key(url):
    """Mask the license key in a URL with stars.

    This function takes a URL that contains embedded basic authentication
    credentials (username and password/license key) and replaces the
    license key with a masked value ("****") for safe logging or display.

    Args:
        url (str):
            The URL containing basic auth credentials.

    Returns:
        str:
            The URL with the license key masked. If no username is present,
            the original URL is returned.
    """
    parsed = urlparse(url)
    if parsed.username:
        netloc = f'{parsed.username}:****@{parsed.hostname}'
        return urlunparse(parsed._replace(netloc=netloc))

    return url


def print_empty_username_or_license_key():
    """Print an error message when username or license key is missing."""
    print_message(
        '\nError installing SDV Enterprise. The provided username and/or license '
        'key was empty. Please make sure to copy-paste your entire username and '
        "license key values. If you're continuing to experience issues, please "
        'reach out to the DataCebo team.'
    )


def build_index_url(username, license_key):
    """Build the authenticated PyPI index URL for package installation.

    This function generates the index URL used to install SDV Enterprise
    packages from a private PyPI server. It encodes the provided username,
    combines it with the license key, and injects them as basic authentication
    credentials in the URL.

    Args:
        username (str):
            The username associated with the license.
        license_key (str):
            The license key required for authentication.

    Returns:
        str:
            A fully formatted index URL that includes the quoted username
            and license key for accessing the private PyPI repository.

    Raises:
        SystemExit:
            If either `username` or `license_key` is empty, the function
            prints an error message and exits the program.
    """
    if not username or not license_key:
        print_empty_username_or_license_key()
        sys.exit(0)

    username_quoted = quote_plus(username, encoding='utf-8')
    return f'https://{username_quoted}:{license_key}@{PYPI_URL}/'



"""System requirements utility functions."""

import platform
import re
import subprocess
import sys
import sysconfig

import pip
from packaging.tags import sys_tags
from packaging.version import Version

from sdv_installer.constants import (
    MAX_PYTHON_VERSION,
    MIN_OS_VERSIONS,
    MIN_PIP_VERSION,
    MIN_PYTHON_VERSION,
    REQUIRED_BITNESS,
    SUPPORTED_ARCHITECTURES,
    SUPPORTED_LINUX_DISTROS,
    SUPPORTED_OSES,
)
from sdv_installer.utils.console_utils import print_message

# IF CHANGING THIS VALUE, UPDATE CAT COMMAND IN get_os_info
LINUX_OS_RELEASE_FILE = '/etc/os-release'


def _normalize_kernel_version(kernel_version):
    parts = kernel_version.split('.')
    normalized_version = []
    for part in parts:
        num = ''
        for ch in part:
            if ch.isdigit():
                num += ch
            else:
                break

        if not num:
            break

        normalized_version.append(num)

    return '.'.join(normalized_version)


def _get_linux_distro_version():
    with open(LINUX_OS_RELEASE_FILE) as f:
        data = dict(line.strip().split('=', 1) for line in f if '=' in line)

    linux_distro_version = data.get('VERSION_ID', '').strip('"')
    return Version(linux_distro_version)


def get_os_info():
    """Detects the operating system type and version.

    Returns:
        Tuple[str, str, Optional[str], bool]: A tuple containing:
            - os_system (str):
                'macOS', 'Linux', or other OS name.
            - os_version (str):
                The complete OS version (e.g., '15.6.1').
            - linux_distro (Optional[str]):
                Pretty name of the Linux distribution, if applicable.
            - is_linux (bool):
                Whether the OS is Linux.
    """
    system = platform.system()
    version = platform.version()
    is_linux = sys.platform == 'linux'
    linux_distro = None

    if sys.platform == 'darwin':
        version = subprocess.check_output(['sw_vers', '-productVersion'], text=True).strip()
        system = 'macOS'

    elif is_linux:
        version = _normalize_kernel_version(platform.release())
        output = subprocess.run('cat /etc/os-release', shell=True, capture_output=True, text=True)
        match = re.search(r'PRETTY_NAME="(.*)"', output.stdout.strip())
        linux_distro = match.group(1) if match else None

    return system, version, linux_distro, is_linux


def get_architecture():
    """Determines the system architecture (e.g., arm64, x86_64).

    Returns:
        str: The system's architecture identifier.
    """
    if platform.system().lower() == 'windows':
        return sysconfig.get_platform()

    return subprocess.check_output(['uname', '-m'], text=True).strip().lower()


def get_current_platform_tag():
    """Returns the normalized platform tag for the current system."""
    return sysconfig.get_platform().replace('-', '_').replace('.', '_')


def get_python_version_code():
    """Returns the current Python major and minor version as a tuple."""
    major = sys.version_info.major
    minor = sys.version_info.minor
    return Version(f'{major}.{minor}')


def get_system_info():
    """Collects all relevant system information for SDV Installer validation.

    Returns:
        dict: A dictionary with system attributes including:
            - os_system (str)
            - os_version (str)
            - linux_distro (Optional[str])
            - architecture (str)
            - python_version (str)
            - system_bit (str)
            - pip_version (str)
            - platform_tag (str)
            - processor (str)
            - is_linux (bool)
    """
    os_system, os_version, linux_distro, is_linux = get_os_info()
    arch = get_architecture()
    python_version = platform.python_version()
    bitness = '64-bit' if sys.maxsize > 2**32 else '32-bit'
    pip_version = pip.__version__
    processor = platform.processor()
    platform_tag = get_current_platform_tag()

    return {
        'os_system': os_system,
        'os_version': os_version,
        'linux_distro': linux_distro,
        'architecture': arch,
        'python_version': python_version,
        'system_bit': bitness,
        'pip_version': pip_version,
        'platform_tag': platform_tag,
        'processor': processor,
        'is_linux': is_linux,
    }


def get_user_supported_tags():
    """Retrieves the set of platform tags supported by the current Python environment.

    These tags are used to determine compatibility with precompiled wheels.

    Returns:
        Set[str]:
            A set of supported platform tag strings.
    """
    return {tag.platform for tag in sys_tags()}


def get_pip_version_code():
    """Retrive the `pip` version of the current environment."""
    return Version(pip.__version__)


def validate_system_requirements(info):
    """Validates the current system against the minimum technical requirements.

    Args:
        info (dict): A dictionary containing system info (as from `get_system_info`).

    Returns:
        List[str]: A list of keys that failed validation (e.g., ['python_version']).
    """
    errors = []

    # Validate Python version
    python_version = get_python_version_code()
    if python_version < MIN_PYTHON_VERSION or python_version > MAX_PYTHON_VERSION:
        errors.append('python_version')

    # Validate system bitness
    if info['system_bit'] != REQUIRED_BITNESS:
        errors.append('system_bit')

    # Validate OS
    if info['os_system'] not in SUPPORTED_OSES:
        errors.append('os_system')

    # Validate os minimal version
    if Version(info['os_version']) < MIN_OS_VERSIONS.get(info['os_system'], Version('0')):
        errors.append('os_version')

    # Validate architecture
    if info['architecture'] not in SUPPORTED_ARCHITECTURES:
        errors.append('architecture')

    # Validate pip minimal version
    if Version(info['pip_version']) < MIN_PIP_VERSION:
        errors.append('pip_version')

    if info.get('linux_distro'):
        linux_distro = info.get('linux_distro', '').lower()
        min_version = None
        linux_version = _get_linux_distro_version()
        if 'ubuntu' in linux_distro:
            min_version = SUPPORTED_LINUX_DISTROS.get('ubuntu')
        elif 'debian' in linux_distro:
            min_version = SUPPORTED_LINUX_DISTROS.get('debian')

        if min_version and linux_version < min_version:
            errors.append('linux_distro')

    return errors


def print_check_results():
    """Runs the system checks and prints the result in a user-friendly format."""
    print_message('Verifying system requirements for SDV Enterprise: \n')
    info = get_system_info()
    errors = validate_system_requirements(info)

    def mark(key):
        return ' (!)' if key in errors else ''

    print_message(f'Operating system: {info["os_system"]}{mark("os_system")}')
    print_message(f'OS Version: {info["os_version"]}{mark("os_version")}')
    if info['is_linux'] and info['linux_distro']:
        print_message(f'Linux Distribution: {info["linux_distro"]}{mark("linux_distro")}')

    print_message(f'Architecture: {info["architecture"]}{mark("architecture")}')
    print_message(f'Python Version: {info["python_version"]}{mark("python_version")}')
    print_message(f'System bit: {info["system_bit"]}{mark("system_bit")}')
    print_message(f'Pip version: {info["pip_version"]}{mark("pip_version")}')

    print_message(f'\nResult: {"Passed" if not errors else "Failed"}')
    if not errors:
        print_message('Your system meets the technical requirements for SDV Enterprise.')
        sys.exit(0)
    else:
        print_message(
            'Your system may not meet the technical requirements for SDV Enterprise. '
            'Please check the items marked by (!) and try updating the SDV '
            "Installer if something doesn't seem right."
        )
        print_message('The most up-to-date requirements are on the SDV Enterprise docs site.')
        sys.exit(1)



"""Installer of SDV-Enterprise."""

import glob
import json
import logging
import os
import pathlib
import shutil
import subprocess
import sys
from typing import Dict, List, Optional, Union

import requests
from platformdirs import user_data_dir
from requests.exceptions import RequestException

from sdv_installer import config
from sdv_installer.authentication import authenticate
from sdv_installer.constants import (
    AUTH_FILE_NAME,
    CONNECTORS_PACKAGES,
    EXPECTED_PACKAGES,
    TRUSTED_HOSTS,
    ExitCode,
    ProductType,
)
from sdv_installer.utils import (
    build_index_url,
    check_is_sdv_enterprise_included,
    determine_additional_sdv_enterprise_deps,
    display_progress_animation,
    get_latest_package_version,
    get_package_name,
    list_current_installed_packages,
    list_current_installed_packages_with_their_version,
    mask_license_key,
    print_additional_dependencies_installed,
    print_failed_to_connect,
    print_invalid_credentials,
    print_message,
    print_package_summary,
    print_warning_base_connector_package_installed,
    read_stored_packages,
    remove_package_name,
    split_base_bundles_and_others,
    store_package_name,
)

LOGGER = logging.getLogger(__name__)


def _perform_pip_action(
    action,
    packages,
    version=None,
    index_url=None,
    extra_options=None,
    show_version=False,
    debug=False,
    upgrade=False,
    package_status=None,
):
    """Performs pip actions (e.g., install, uninstall) on a list of packages.

    This function builds and executes pip commands using subprocess for each package
    provided. It supports custom package indexes and additional pip options.

    Args:
        action (str):
            The pip action to perform (e.g., 'install', 'uninstall').
        packages (list[str]):
            A list of package names to process.
        index_url (str, optional):
            Custom PyPI index URL. Defaults to None.
        extra_options (list[str], optional):
            Additional pip CLI options. Defaults to None.
        debug (bool, optional):
            If True, prints pip commands instead of suppressing output.
            Defaults to False.
    """
    package_status = package_status or {}
    extra_options = extra_options or []
    installed_packages_with_versions = list_current_installed_packages_with_their_version()
    package_version = None
    was_successful = False

    for package_name in packages:
        package = package_name
        use_version = bool(version or index_url) and package not in CONNECTORS_PACKAGES
        online_access = '--no-index' not in extra_options

        if use_version and online_access:
            package_version = version or get_latest_package_version(package_name, index_url)
            if upgrade is False:
                package = f'{package_name}=={package_version}' if package_version else package_name

        elif package in CONNECTORS_PACKAGES and online_access:
            package_version = get_latest_package_version(package_name, index_url)

        command = f'{sys.executable} -m pip {action} {package or package_name}'
        printable_command = f'\npip {action} {package or package_name}'
        if index_url:
            command = f'{command} --index-url {index_url}'
            printable_command = f'{printable_command} --index-url {mask_license_key(index_url)}'
            trusted_hosts_str = ' '.join(f'--trusted-host {host}' for host in TRUSTED_HOSTS)
            command = f'{command} {trusted_hosts_str}'
            printable_command = f'{printable_command} {trusted_hosts_str}'

        if upgrade:
            command = f'{command} --upgrade'
            printable_command = f'{printable_command} --upgrade'

        if extra_options:
            command = f'{command} {" ".join(extra_options)}'
            printable_command = f'{printable_command} {" ".join(extra_options)}'

        LOGGER.debug('Launching subprocess with: %s', printable_command)
        if debug:
            print_message(printable_command)
            process = subprocess.run(command.split())

        else:
            process = subprocess.Popen(
                command.split(),
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            display_progress_animation(
                process=process,
                package=package_name,
                version=package_version,
                action=action,
                upgrade=upgrade,
                installed_packages=installed_packages_with_versions,
                show_version=show_version,
            )

        was_successful = process.returncode == ExitCode.SUCCESS
        package_status[package_name] = was_successful

        if action == 'install' and not extra_options and was_successful:
            store_package_name(package_name)

        if 'uninstall' in action:
            remove_package_name(package_name)

    return package_status


def _print_and_perform_action(
    action: str,
    packages: Dict[str, str],
    index_url: Union[str, None],
    debug: bool,
    version: Union[str, None],
    pip_options: Optional[List[str]] = None,
    upgrade: bool = False,
    use_product_type: bool = True,
):
    """Print and perform pip action (install/download) for base and bundle packages.

    Args:
        action (str):
            The pip action to perform (eg. 'install' or 'download').
        packages (dict):
            A dictionary mapping package name to the product type.
        index_url (str):
            The index url to use for pip.
        debug (bool):
            Whether or not to show full print out from pip.
        version (str):
            The version of the package to do the action on.
        pip_options (list):
            List of pip options (eg. --python-version).
        upgrade (bool):
            Whether or not to use upgrade option with pip.
        use_product_type (bool):
            Whether or not to use the product type to aid in messages that get printed.
    """
    base_packages, bundles, others = split_base_bundles_and_others(packages, use_product_type)
    action_cap = action.capitalize()
    package_status = {}

    if base_packages:
        print_message(f'\n{action_cap}ing SDV:')
        base_package_statuses = _perform_pip_action(
            action=action,
            packages=base_packages,
            index_url=index_url,
            debug=debug,
            version=version,
            extra_options=pip_options,
            show_version=True,
            upgrade=upgrade,
            package_status=package_status,
        )
        package_status.update(base_package_statuses)

    if bundles:
        print_message(f'\n{action_cap}ing Bundles:')
        bundle_package_statuses = _perform_pip_action(
            action=action,
            packages=bundles,
            index_url=index_url,
            debug=debug,
            version=version,
            extra_options=pip_options,
            show_version=False,
            upgrade=upgrade,
        )
        package_status.update(bundle_package_statuses)

    if others:
        print_message(f'\n{action_cap}ing Packages:')
        other_package_statuses = _perform_pip_action(
            action=action,
            packages=others,
            index_url=index_url,
            debug=debug,
            version=version,
            extra_options=pip_options,
            show_version=False,
            upgrade=upgrade,
        )
        package_status.update(other_package_statuses)

    if set(CONNECTORS_PACKAGES).intersection(packages):
        print_warning_base_connector_package_installed()

    # Determine if sdv-enterprise is actually present
    if action == 'install':
        successful_packages = list_current_installed_packages()
    else:
        successful_packages = [pkg for pkg, status in package_status.items() if status]

    sdv_included = check_is_sdv_enterprise_included(successful_packages)
    print_package_summary(action, package_status, sdv_included)


def _uninstall_packages_action(packages, debug=False):
    """Uninstall the packages using subprocess / Popen with ``pip``."""
    print_message('Uninstalling:')
    _perform_pip_action('uninstall -y', packages, index_url=None, debug=debug)


def _install_packages_action(
    username, license_key, user_requested_packages, version=None, debug=False, upgrade=False
):
    """Install the packages using subprocess / Popen with ``pip``."""
    index_url = build_index_url(username, license_key)
    _print_and_perform_action(
        action='install',
        packages=user_requested_packages,
        index_url=index_url,
        debug=debug,
        version=version,
        pip_options=None,
        upgrade=upgrade,
    )


def _download_packages_action(
    username, license_key, packages, python_version, platform, folder, version=None, debug=False
):
    """Download the packages using subprocess / Popen with ``pip download``."""
    index_url = build_index_url(username, license_key)

    pip_options = [
        '--dest',
        folder,
    ]
    if platform:
        pip_options.append('--only-binary=:all:')
        for item in platform:
            pip_options.extend(['--platform', item])

    if python_version:
        if '--only-binary=:all:' not in pip_options:
            pip_options.append('--only-binary=:all:')

        pip_options.extend(['--python-version', python_version])

    _print_and_perform_action(
        action='download',
        packages=packages,
        index_url=index_url,
        debug=debug,
        version=version,
        pip_options=pip_options,
    )


def _get_accessible_packages(username, license_key):
    """Get a dictionary mapping all packages the user can access to their product type."""
    post_data = {'username': username, 'license_key': license_key}

    try:
        response = requests.post(
            config.API_PACKAGE_PERMISSIONS,
            data=json.dumps(post_data),
            headers=config.HEADERS,
            timeout=config.TIMEOUT,
        )
    except RequestException:
        print_failed_to_connect()
        return None

    if response.status_code != requests.codes.ok:
        print_invalid_credentials()
        return None

    parsed_response = response.json()
    LOGGER.debug('Active packages user %s has access to: %s', username, parsed_response)
    accessible_packages = parsed_response.get('packages', [])
    package_to_product_type = {
        pkg['package_name']: pkg['product_type'] for pkg in accessible_packages
    }
    return package_to_product_type


def _get_packages(username, license_key, package=None, options=None):
    """Retrieve a list of available packages to install."""
    package_to_product_type = _get_accessible_packages(username, license_key)
    if not package_to_product_type:
        return

    if package in package_to_product_type and options:
        options = options if isinstance(options, list) else [options]
        bundle_available_packages = {
            f'{package}[{option}]': ProductType.BUNDLE.value for option in options
        }
        return bundle_available_packages

    elif package and package not in package_to_product_type:
        print_message(f'Invalid package {package} or no access to the given package.')
        return

    if package and package in package_to_product_type:
        return {package: package_to_product_type[package]}

    return package_to_product_type


def _create_user_data_dir(data_dir, error_message=None):
    if not error_message:
        error_message = "Unable to write auth file to user's data directory. Permissions required."
    try:
        data_dir_path = pathlib.Path(data_dir)
        data_dir_path.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        raise RuntimeError(error_message)


def _save_auth_file(username, license_key, folder=None):
    """Retrieve the auth file for a user and save it to their appdata folder.

    Args:
        username (str):
            Username of user to get auth file for.
        license_key (str):
            User's license key.
        folder (str | None):
            Folder to save the auth file to.
    """
    error_message = None
    if folder:
        data_dir = folder
        error_message = 'Unable to create specified folder. Permissions required.'
    else:
        data_dir = user_data_dir(appname='sdv-enterprise', appauthor='Datacebo')

    _create_user_data_dir(data_dir, error_message)
    auth_file_path = pathlib.Path(data_dir, AUTH_FILE_NAME)
    try:
        response = requests.post(
            config.API_AUTH_FILE,
            data=json.dumps({'username': username, 'license_key': license_key}),
            headers=config.HEADERS,
            timeout=config.TIMEOUT,
        )
        response.raise_for_status()
    except requests.HTTPError as e:
        raise RuntimeError('Could not fetch auth file') from e

    auth_json = response.json()
    with open(auth_file_path, 'w') as f:
        json.dump(auth_json, f)


def _move_auth_file(source_folder, destination=None):
    """Move the auth file to the required location.

    Args:
        source_folder (str):
            The folder the auth file is in.
        destination (str | None):
            The path to move the file to. If None, then move to the user data dir.
    """
    if destination is None:
        destination = user_data_dir(appname='sdv-enterprise', appauthor='Datacebo')
        _create_user_data_dir(destination)

    auth_file_path = pathlib.Path(source_folder, AUTH_FILE_NAME)
    if os.path.exists(auth_file_path):
        shutil.copy(auth_file_path, pathlib.Path(destination) / AUTH_FILE_NAME)


@authenticate
def get_auth_file(username=None, license_key=None, folder=None, **kwargs):
    """Retrieve the auth file from the license key service.

    Args:
        username (str):
            The username for SDV Enterprise.
        license_key (str):
            The license key.
        folder (str):
            The folder path to save the auth file to. If None, it will save to the
            location SDV Enterprise requires.
    """
    _save_auth_file(username=username, license_key=license_key, folder=folder)


@authenticate
def install_packages(
    username=None,
    license_key=None,
    package=None,
    options=None,
    debug=False,
    version=None,
    upgrade=False,
    **kwargs,
):
    """Install packages in the current environment (online)."""
    _save_auth_file(username, license_key)
    all_sdv_enterprise_related_pkgs = _get_accessible_packages(username, license_key)
    packages_before_install = list_current_installed_packages()

    user_requested_packages = _get_packages(
        username=username,
        license_key=license_key,
        package=package,
        options=options,
    )
    if not user_requested_packages:
        return

    _install_packages_action(
        username,
        license_key,
        user_requested_packages=user_requested_packages,
        version=version,
        debug=debug,
        upgrade=upgrade,
    )
    packages_after_install = list_current_installed_packages()
    additional_dependencies_installed = determine_additional_sdv_enterprise_deps(
        packages_before_install=packages_before_install,
        packages_after_install=packages_after_install,
        user_requested_packages=user_requested_packages,
        all_sdv_enterprise_related_pkgs=all_sdv_enterprise_related_pkgs,
    )
    print_additional_dependencies_installed(additional_dependencies_installed)


def install_packages_from_folder(folder, package=None, options=None, debug=False):
    """Install packages from a folder (offline)."""
    wheels = glob.glob(f'{folder}/*.whl')
    _move_auth_file(folder)
    packages = set()
    for wheel in wheels:
        package_name = get_package_name(wheel)
        if package_name in EXPECTED_PACKAGES:
            packages.add(package_name)

    if package:
        if package not in packages:
            print_message(
                f"\nWarning: We couldn’t find the requested package: '{package}' in the provided "
                'folder, no packages were installed.'
            )
            return
        else:
            packages = [package]
            if options:
                packages = [f'{package}[{option}]' for option in options]

    pip_options = ['--no-index', f'--find-links {folder}']
    packages_before_install = list_current_installed_packages()
    packages = dict.fromkeys(packages)

    _print_and_perform_action(
        'install',
        packages=packages,
        pip_options=pip_options,
        version=None,
        index_url=None,
        debug=debug,
        use_product_type=False,
    )
    packages_after_install = list_current_installed_packages()

    additional_dependencies_installed = determine_additional_sdv_enterprise_deps(
        packages_before_install=packages_before_install,
        packages_after_install=packages_after_install,
        user_requested_packages=package,
        all_sdv_enterprise_related_pkgs=EXPECTED_PACKAGES,
    )
    print_additional_dependencies_installed(additional_dependencies_installed)


@authenticate
def _auth_get_packages(username=None, license_key=None, package=None, **kwargs):
    """Wrapper around `_get_packages` to authenticate before using it."""
    return _get_packages(username=username, license_key=license_key, package=package, **kwargs)


def uninstall_packages(
    all_packages=False,
    debug=False,
    **kwargs,
):
    """Uninstall previously installed packages or all accessible packages."""
    packages = []
    try:
        packages = read_stored_packages()

    except (FileNotFoundError, PermissionError, OSError, ValueError):
        print_message('Fetching package list from license server...')
        packages = _auth_get_packages()

    if not packages:
        print_message('No packages found to uninstall.')
        return

    current_packages = list_current_installed_packages()
    if set(packages).intersection(set(current_packages)):
        _uninstall_packages_action(packages, debug=debug)
    else:
        print_message('No packages found to uninstall.')


@authenticate
def download_packages(
    username=None,
    license_key=None,
    package=None,
    python_version=None,
    platform=None,
    folder=None,
    version=None,
    options=None,
    debug=False,
    **kwargs,
):
    """Download packages into a local folder for offline installation."""
    _save_auth_file(username, license_key, folder=folder)
    packages = _get_packages(
        username=username, license_key=license_key, package=package, options=options
    )

    if not packages:
        print_message('No packages available for download.')
        return

    _download_packages_action(
        username=username,
        license_key=license_key,
        packages=packages,
        python_version=python_version,
        platform=platform,
        folder=folder,
        version=version,
        debug=debug,
    )


@authenticate
def list_packages(username: str, license_key: str, **kwargs):
    """List all packages the user has access to."""
    package_to_product_type = _get_accessible_packages(username, license_key)

    # Filter bundles
    base_packages, bundles, others = split_base_bundles_and_others(package_to_product_type)
    bases_message = '\n'.join(base_packages)
    print_message(f'\nSDV Enterprise:\n{bases_message}')
    if bundles:
        bundles_message = '\n'.join(bundles)
        print_message(f'\nSDV Bundles:\n{bundles_message}')

    if others:
        others_message = '\n'.join(others)
        print_message(f'\nAdditional Packages:\n{others_message}')

    return package_to_product_type



"""Utility functions for package and virtual env."""

import os
import pathlib
import subprocess
import sys
from importlib.metadata import distributions
from typing import Dict, Tuple

from packaging.requirements import Requirement
from packaging.utils import parse_sdist_filename, parse_wheel_filename
from packaging.version import InvalidVersion, Version

from sdv_installer.config import TIMEOUT
from sdv_installer.constants import EXPECTED_PACKAGES, ProductType

BASE_PREFIX = 'sdv-enterprise'
BUNDLE_PREFIX = 'bundle-'


def list_current_installed_packages():
    """Returns a list of currently installed package names.

    Uses `importlib.metadata.distributions()` to retrieve metadata for all installed
    distributions and extracts their package names.

    Returns:
        List[str]:
            A list of package names currently installed in the environment.
    """
    return [dist.metadata['Name'] for dist in distributions()]


def list_current_installed_packages_with_their_version():
    """Returns a list of currently installed package names.

    Uses `importlib.metadata.distributions()` to retrieve metadata for all installed
    distributions and extracts their package names.

    Returns:
        List[str]:
            A list of package names currently installed in the environment.
    """
    return {dist.metadata['Name']: dist.metadata['Version'] for dist in distributions()}


def clean_path(path: str):
    """Cleans a file path by stripping leading/trailing whitespace and converting to lowercase.

    Args:
        path (str):
            The input file path as a string.

    Returns:
        str:
            The cleaned file path.
    """
    return path.strip().lower()


def get_package_name(file_path: str):
    """Extracts the package name from a given file path.

    Supports `.whl`, `.tar.gz`, and other archive formats. For wheel and source distribution
    files, the name is parsed using appropriate helper functions. For other file types,
    the base file name is returned.

    Args:
        file_path (str):
            Path to the package file.

    Returns:
        str or None:
            The extracted package name, or None if the path is invalid.
    """
    package_name = None
    if not file_path or not file_path.strip():
        return None

    file_path = clean_path(file_path)
    if file_path.endswith('.whl') or file_path.endswith('.tar.gz'):
        filename = os.path.basename(file_path)
        if file_path.endswith('.whl'):
            package_name, _, _, _ = parse_wheel_filename(filename)
        else:
            package_name, _ = parse_sdist_filename(filename)
    else:
        package_name = pathlib.PurePath(file_path).name

    return package_name


def get_latest_package_version(package_name, index_url=None):
    """Returns a sorted list of available versions for a given package.

    This function invokes `pip index versions` via subprocess to retrieve the
    list of available versions for a specified package from the given Python
    package index (PyPI or a custom index).

    Args:
        package_name (str):
            The name of the package to query.
        index_url (str, optional):
            The base URL of the package index to use. If not provided, the default PyPI
            index is used.

    Returns:
        str:
            Returns a string representing the highest version available.
            Returns None if the pip command fails.
    """
    cmd = [sys.executable, '-m', 'pip', 'index', 'versions', package_name]
    if index_url:
        cmd += ['--index-url', index_url]

    result = subprocess.run(cmd, capture_output=True, text=True, timeout=TIMEOUT)
    if result.returncode != 0:
        return None

    for line in result.stdout.splitlines():
        if line.startswith('Available versions: '):
            text_versions = line.replace('Available versions: ', '')
            text_versions = text_versions.split(',')
            break

    return text_versions[0]


def split_base_bundles_and_others(
    packages: Dict[str, str], use_product_type: bool = True
) -> Tuple[str, list]:
    """Split base sdv-enterprise packages from bundles.

    Args:
        packages (dict):
            List of package dicts. Each dict has the keys 'package_name' and 'package_type'.
        use_product_type (bool):
            Whether or not to use the product type attribute to figure out
            if a package is a base package or bundle. If false, fall back to using the package name.

    Returns:
        tuple (list[str], list[str]). A list of base packages and a list of bundle packages.
    """
    base_packages = []
    bundles = []
    others = []
    if use_product_type:
        for package, product_type in packages.items():
            if product_type == ProductType.BASE.value:
                base_packages.append(package)
            elif product_type == ProductType.BUNDLE.value:
                bundles.append(package)
            else:
                others.append(package)

    else:
        for package in packages:
            if package.startswith(BASE_PREFIX):
                base_packages.append(package)
            elif package.startswith(BUNDLE_PREFIX):
                bundles.append(package)
            else:
                others.append(package)

    return sorted(base_packages), sorted(bundles), sorted(others)


def is_version_bigger(installed_version, requested_version):
    """Check if the installed version is greater than the requested version."""
    if isinstance(installed_version, str) and isinstance(requested_version, str):
        try:
            return Version(installed_version) > Version(requested_version)
        except InvalidVersion:
            return False

    return False


def is_version_equal(installed_version, requested_version):
    """Check if the installed version is equal to the requested version."""
    if isinstance(installed_version, str) and isinstance(requested_version, str):
        try:
            return Version(installed_version) == Version(requested_version)
        except InvalidVersion:
            return False

    return False


def convert_to_set(obj):
    """Convert a list, tuple, str, int, float, or dictionary to a set."""
    if isinstance(obj, (list, tuple)):
        obj = set(obj)
    elif isinstance(obj, (str, int, float)):
        obj = {obj}
    elif isinstance(obj, dict):
        obj = set(obj.keys())
    return obj


def get_requirement_name(requirement_name):
    """Get the name of a requirement, useful to remove extra_requires, specifiers, markers."""
    if isinstance(requirement_name, str):
        return Requirement(requirement_name).name
    return requirement_name


def is_none_or_empty_iterate(obj):
    """Check if object is None or an empty list/dict/iterate."""
    return obj is None or (hasattr(obj, '__len__') and len(obj) == 0)


def determine_additional_sdv_enterprise_deps(
    packages_before_install,
    packages_after_install,
    user_requested_packages,
    all_sdv_enterprise_related_pkgs=EXPECTED_PACKAGES,
):
    """Determine the additional sdv-enterprise related packages installed, given before & after.

    Args:
        packages_before_install (set, list[str]):
            All packages before installation with sdv-installer.

        packages_after_install (set, list[str]):
            All packages after installation with sdv-installer.

        user_requested_packages (set, list[str], dict[str, str], str, None)
            The user requested packages to install. If it is an itert

        all_sdv_enterprise_related_pkgs (set, list[str])
            The package names of all base packages and bundle packages
            that could be installed by the user.
            Defaults to EXPECTED_PACKAGES

    Returns:
        (set): The additional sdv-enterprise related packages that were installed.
    """
    if is_none_or_empty_iterate(user_requested_packages):
        # User did not say a specific package to install
        # Return empty list to specify no additional deps were installed (everything installed)
        return {}

    packages_before_install = convert_to_set(packages_before_install)
    packages_after_install = convert_to_set(packages_after_install)

    user_requested_packages = convert_to_set(user_requested_packages)
    user_requested_packages = {get_requirement_name(pkg) for pkg in user_requested_packages}

    all_sdv_enterprise_related_pkgs = convert_to_set(all_sdv_enterprise_related_pkgs)

    all_installed_pkgs = packages_after_install.difference(packages_before_install)
    installed_pkgs_related_to_sdv_enterprise = all_installed_pkgs.intersection(
        all_sdv_enterprise_related_pkgs
    )
    additional_dependencies_installed = installed_pkgs_related_to_sdv_enterprise.difference(
        user_requested_packages
    )
    return additional_dependencies_installed


def check_is_sdv_enterprise_included(packages):
    """Check if any of the given package names includes 'sdv-enterprise'.

    Args:
        packages (list of str):
            A list of package names.

    Returns:
        bool:
            True if 'sdv-enterprise' is found in any package name, False otherwise.
    """
    for package in packages:
        if package.startswith(BASE_PREFIX):
            return True

    return False



"""Utilities to handle appdirs and store application data."""

import logging
import os
from pathlib import Path

from platformdirs import user_data_dir

APP_NAME = 'SDV-Installer'
APP_AUTHOR = 'DataCebo'
LOGGER = logging.getLogger(__name__)
CONFIG_PATH = Path(__file__).resolve().parent


def store_package_name(package):
    """Store the installed package name in a text file in a platform-specific directory.

    Args:
        packages (str):
            A string representing the package name.
    """
    storage_dir = Path(user_data_dir(APP_NAME, APP_AUTHOR))

    try:
        storage_dir.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        # No access to user data dir, fall back if possible
        if os.access(CONFIG_PATH, os.W_OK):
            storage_dir = CONFIG_PATH
        else:
            # Nowhere to write
            return

    except OSError:
        # Filesystem might be read-only or inaccessible
        return

    package_file = storage_dir / 'installed_packages.txt'
    try:
        with package_file.open('a') as installed_packages:
            installed_packages.write(package + '\n')

    except (PermissionError, OSError) as error:
        LOGGER.error(f"Failed to write installed package '{package}' to '{package_file}': {error}")


def read_stored_packages():
    """Read the list of installed packages from the platform-specific data file.

    Returns:
        list[str]:
            A list of installed package names.

    Raises:
        FileNotFoundError:
            If the file does not exist.
    """
    storage_dir = Path(user_data_dir(APP_NAME, APP_AUTHOR))
    package_file = storage_dir / 'installed_packages.txt'

    if not package_file.exists():
        raise FileNotFoundError('No stored package file found.')

    with package_file.open('r') as installed_packages:
        return set([line.strip() for line in installed_packages if line.strip()])


def remove_package_name(package):
    """Remove a single package name from the installed packages file.

    Args:
        package (str):
            A string representing the package name to remove.
    """
    storage_dir = Path(user_data_dir(APP_NAME, APP_AUTHOR))
    package_file = storage_dir / 'installed_packages.txt'

    if not package_file.exists():
        return

    # Read all packages and filter out the one to remove
    with package_file.open('r') as package_reader:
        packages = [line.strip() for line in package_reader if line.strip() != package]

    with package_file.open('w') as package_writer:
        package_writer.write('\n'.join(packages) if packages else '')



"""Utility module."""

from sdv_installer.utils.data_storage import (
    read_stored_packages,
    remove_package_name,
    store_package_name,
)

from sdv_installer.utils.console_utils import (
    build_index_url,
    display_progress_animation,
    get_password_input,
    mask_license_key,
    print_additional_dependencies_installed,
    print_empty_username_or_license_key,
    print_failed_to_connect,
    print_invalid_credentials,
    print_message,
    print_package_summary,
    print_warning_base_connector_package_installed,
)
from sdv_installer.utils.package_utils import (
    check_is_sdv_enterprise_included,
    get_latest_package_version,
    get_package_name,
    list_current_installed_packages,
    list_current_installed_packages_with_their_version,
    split_base_bundles_and_others,
    determine_additional_sdv_enterprise_deps,
)
from sdv_installer.utils.request_error_handling import handle_http_error_response

__all__ = (
    'build_index_url',
    'check_is_sdv_enterprise_included',
    'display_progress_animation',
    'determine_additional_sdv_enterprise_deps',
    'get_latest_package_version',
    'get_package_name',
    'get_password_input',
    'handle_http_error_response',
    'list_current_installed_packages',
    'list_current_installed_packages_with_their_version',
    'mask_license_key',
    'print_additional_dependencies_installed',
    'print_empty_username_or_license_key',
    'print_failed_to_connect',
    'print_invalid_credentials',
    'print_message',
    'print_package_summary',
    'print_warning_base_connector_package_installed',
    'read_stored_packages',
    'remove_package_name',
    'split_base_bundles_and_others',
    'store_package_name',
)
